# This file was auto-generated by Fern from our API Definition.

import typing
from .chat_completion_assistant_message_param_content_item import ChatCompletionAssistantMessageParamContentItem

ChatCompletionAssistantMessageParamContent = typing.Union[
    str, typing.List[ChatCompletionAssistantMessageParamContentItem]
]
