# This file was auto-generated by Fern from our API Definition.

import typing
from .omitted_reasoning_content import OmittedReasoningContent
from .reasoning_content import ReasoningContent
from .redacted_reasoning_content import RedactedReasoningContent
from .text_content import TextContent
from .tool_call_content import ToolCallContent
from .tool_return_content import ToolReturnContent

MessageContentItem = typing.Union[
    OmittedReasoningContent, ReasoningContent, RedactedReasoningContent, TextContent, ToolCallContent, ToolReturnContent
]
