# This file was auto-generated by Fern from our API Definition.

from .create_voice_chat_completions_request_body import CreateVoiceChatCompletionsRequestBody

__all__ = ["CreateVoiceChatCompletionsRequestBody"]
