# This file was auto-generated by Fern from our API Definition.

import typing
from .system_message import SystemMessage
from .user_message import UserMessage
from .reasoning_message import ReasoningMessage
from .hidden_reasoning_message import HiddenReasoningMessage
from .tool_call_message import ToolCallMessage
from .tool_return_message import ToolReturnMessage
from .assistant_message import AssistantMessage

LettaMessageUnion = typing.Union[
    SystemMessage,
    UserMessage,
    ReasoningMessage,
    HiddenReasoningMessage,
    ToolCallMessage,
    ToolReturnMessage,
    AssistantMessage,
]
